/*
   vmdk-loop: Allow mounting VMware virtual disks via nbd
   Copyright (C) Matthew Chapman 2003

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "vmdk-loop.h"

int disk_fd, nbd;
struct cowdisk_header header;
off_t disk_base, disk_limit;
unsigned int granule_size;
uint32_t l1dir[L1_SIZE];

unsigned int cached_l2dir;
uint32_t l2dir[L2_SIZE];

static size_t read_physical(off_t offset, size_t length, void *buffer)
{
	size_t n;

	if (lseek(disk_fd, offset, SEEK_SET) == -1)
	{
		perror("lseek");
		return -1;
	}

	n = read(disk_fd, buffer, length);
	if (n == -1)
	{
		perror("read from disk");
		return -1;
	}

	return n;
}

static size_t write_physical(off_t offset, size_t length, void *buffer)
{
	size_t n;

	if (lseek(disk_fd, offset, SEEK_SET) == -1)
	{
		perror("lseek");
		return -1;
	}

	n = write(disk_fd, buffer, length);
	if (n == -1)
	{
		perror("write to disk");
		return -1;
	}

	return n;
}

static unsigned int allocate_physical(unsigned int nsectors)
{
	unsigned int sector = header.file_sectors;

	header.file_sectors += nsectors;
	write_physical(0, sizeof(header), (char *)&header);
	ftruncate(disk_fd, header.file_sectors << SECTOR_BITS);

	return sector;
}

static size_t read_virtual(off_t offset, size_t length, void *buffer)
{
	unsigned int granule_index, granule_offset;
	unsigned int l1index, l2index;

	granule_index = offset / granule_size;
	granule_offset = offset % granule_size;
	length = MIN(length, granule_size - granule_offset);

	l1index = (granule_index >> L2_BITS) & L1_MASK;
	l2index = granule_index & L2_MASK;

	if (l1dir[l1index] == 0)
		goto zero_fill;

	else if (l1index != cached_l2dir)
	{
		if (read_physical(l1dir[l1index] << SECTOR_BITS, sizeof(l2dir), (char *)l2dir) != sizeof(l2dir))
			 return 0;

		cached_l2dir = l1index;
	}

	if (l2dir[l2index] == 0)
		goto zero_fill;

	if (read_physical((l2dir[l2index] << SECTOR_BITS) + granule_offset, length, buffer) != length)
		return 0;

	return length;

zero_fill:
	memset(buffer, 0, length);
	return length;
}

static size_t write_virtual(off_t offset, size_t length, void *buffer)
{
	unsigned int granule_index, granule_offset;
	unsigned int l1index, l2index;
	unsigned int block;

	granule_index = offset / granule_size;
	granule_offset = offset % granule_size;
	length = MIN(length, granule_size - granule_offset);

	l1index = (granule_index >> L2_BITS) & L1_MASK;
	l2index = granule_index & L2_MASK;

	if (l1dir[l1index] == 0)
	{
		block = allocate_physical(sizeof(l2dir) >> SECTOR_BITS);
		if (block == 0)
		{
			printf("allocate failed\n");
			return 0;
		}

		l1dir[l1index] = block;
		write_physical(header.l1dir_sector << SECTOR_BITS, sizeof(l1dir), (char *)l1dir);
		
		memset(l2dir, 0, sizeof(l2dir));
		cached_l2dir = l1index;
	}

	else if (l1index != cached_l2dir)
	{
		if (read_physical(l1dir[l1index] << SECTOR_BITS, sizeof(l2dir), (char *)l2dir) != sizeof(l2dir))
			 return 0;

		cached_l2dir = l1index;
	}

	if (l2dir[l2index] == 0)
	{
		block = allocate_physical(header.granularity);
		if (block == 0)
		{
			printf("allocate failed\n");
			return 0;
		}

		l2dir[l2index] = block;
		write_physical(l1dir[l1index] << SECTOR_BITS, sizeof(l2dir), (char *)l2dir);
	}

	if (write_physical((l2dir[l2index] << SECTOR_BITS) + granule_offset, length, buffer) != length)
	{
		printf("write_physical failed\n");
		return 0;
	}

	return length;
}

static void read_request(off_t offset, size_t length, int client)
{
	uint8_t buffer[65536];
	size_t readlength, thislength;
	unsigned int buf_offset;
	int doit = 1;

	if (offset > disk_limit)
	{
		fprintf(stderr, "[vmdk-loop] Read past end of disk: offset %llx\n", (unsigned long long)offset);
		doit = 0;
	}

	while (length > 0)
	{
		readlength = MIN(length, sizeof(buffer));
		if (doit)
			readlength = read_virtual(offset, readlength, buffer);
		else
			memset(buffer, 0, readlength);

		if (readlength == 0)
		{
			fprintf(stderr, "[vmdk-loop] Read from disk failed: offset %llx\n",
				(unsigned long long)offset);
			doit = 0;
			continue;
		}

		offset += readlength;
		length -= readlength;
		buf_offset = 0;
		while (readlength > 0)
		{
			thislength = write(client, buffer + buf_offset, readlength);
			if (thislength <= 0)
			{
				perror("write to nbd");
				return;
			}

			buf_offset += thislength;
			readlength -= thislength;
		}
	}
}

static void write_request(off_t offset, size_t length, int client)
{
	uint8_t buffer[65536];
	size_t readlength, thislength;
	unsigned int buf_offset;
	int doit = 1;

	if (offset > disk_limit)
	{
		fprintf(stderr, "[vmdk-loop] Write past end of disk: offset %llx\n", (unsigned long long)offset);
		doit = 0;
	}

	while (length > 0)
	{
		readlength = MIN(length, sizeof(buffer));
		readlength = read(client, buffer, readlength);
		if (readlength <= 0)
		{
			perror("read from nbd");
			return;
		}

		length -= readlength;
		buf_offset = 0;
		while (readlength > 0)
		{
			if (doit)
				thislength = write_virtual(offset, readlength, buffer + buf_offset);
			else
				thislength = readlength;

			if (thislength == 0)
			{
				fprintf(stderr, "[vmdk-loop] Write to disk failed: offset %llx\n",
					(unsigned long long)offset);
				doit = 0;
				continue;
			}

			offset += thislength;
			buf_offset += thislength;
			readlength -= thislength;
		}
	}
}

int open_disk(char *filename, int allow_write)
{
	disk_fd = open(filename, allow_write ? O_RDWR : O_RDONLY);
	if (disk_fd == -1)
	{
		perror(filename);
		return -1;
	}

	if (read_physical(0, sizeof(header), (char *)&header) != sizeof(header))
		return -1;

	if (memcmp(header.magic, "COWD", 4) != 0)
	{
		fprintf(stderr, "[vmdk-loop] %s is not a VMware virtual disk image\n", filename);
		return -1;
	}

	granule_size = header.granularity << SECTOR_BITS;
	if (read_physical(header.l1dir_sector << SECTOR_BITS, sizeof(l1dir), (char *)l1dir) != sizeof(l1dir))
		return -1;

	cached_l2dir = -1;
	return 0;
}

int read_partition_info(unsigned int partition)
{
	uint8_t sector0[SECTOR_SIZE];
	struct partition_info *info;

	if (partition == 0)
	{
		disk_base = 0;
		disk_limit = header.disk_sectors << SECTOR_BITS;
		return 0;
	}

	if (partition > 4)
	{
		fprintf(stderr, "Error: Only primary partitions supported currently\n");
		return -1;
	}

	read_virtual(0, SECTOR_SIZE, sector0);
	if ((sector0[SECTOR_SIZE-2] != 0x55) || (sector0[SECTOR_SIZE-1] != 0xaa))
	{
		fprintf(stderr, "Error: No PC partition table on drive\n");
		return -1;
	}

	info = (struct partition_info *)(sector0 + SECTOR_SIZE - 2) - (5 - partition);
	if (info->size == 0)
	{
		fprintf(stderr, "Error: partition %d does not exist\n", partition);
		return -1;
	}

	printf("start = %d, size = %d\n", info->start, info->size);
	disk_base = info->start << SECTOR_BITS;
	disk_limit = (info->start + info->size) << SECTOR_BITS;
	return 0;
}

int connect_nbd(char *nbdev, unsigned long devsize)
{
	int sock[2];

#ifdef USE_UNIX_SOCKETS
	if (socketpair(PF_UNIX, SOCK_STREAM, 0, sock) == -1)
	{
		perror("socketpair");
		return -1;
	}
#else /* Linux 2.4 TCP doesn't support socketpair */
	struct sockaddr_in addr;
	socklen_t addrlen = sizeof(addr);
	int client;

	sock[0] = socket(PF_INET, SOCK_STREAM, 0);
	sock[1] = socket(PF_INET, SOCK_STREAM, 0);
	if ((sock[0] == -1) || (sock[1] == -1))
	{
		perror("socket");
		return -1;
	}

	addr.sin_family = AF_INET;
	addr.sin_port = 0;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sock[1], (struct sockaddr *)&addr, sizeof(addr)) == -1)
	{
		perror("bind");
		return -1;
	}

	if (getsockname(sock[1], (struct sockaddr *)&addr, &addrlen) == -1)
	{
		perror("getsockname");
		return -1;
	}

	listen(sock[1], 1);
#endif

	nbd = open(nbdev, O_RDWR);
	if (nbd == -1)
	{
		perror(nbdev);
		return -1;
	}

	ioctl(nbd, NBD_CLEAR_SOCK);
	if (ioctl(nbd, NBD_SET_SOCK, sock[0]) == -1)
	{
		perror("NBD_SET_SOCK");
		goto error;
	}

	if (ioctl(nbd, NBD_SET_BLKSIZE, SECTOR_SIZE) == -1)
	{
		perror("NBD_SET_BLKSIZE");
		goto error;
	}

	if (ioctl(nbd, NBD_SET_SIZE_BLOCKS, devsize >> SECTOR_BITS) == -1)
	{
		perror("NBD_SET_SIZE_BLOCKS");
		goto error;
	}

	switch (fork())
	{
		case -1:
			perror("fork");
			goto error;

#ifdef USE_UNIX_SOCKETS
		case 0: /* child */
			close(sock[1]);
			break;

		default: /* parent */
			close(sock[0]);
			return sock[1];
#else
		case 0: /* child */
			close(sock[1]);
			if (connect(sock[0], (struct sockaddr *)&addr, sizeof(addr)) == -1)
			{
				perror("connect");
				goto error;
			}
			break;

		default: /* parent */
			close(sock[0]);
			client = accept(sock[1], NULL, 0);
			close(sock[1]);
			if (client == -1)
				perror("accept");
			return client;
#endif
	}

	if (ioctl(nbd, NBD_DO_IT) == -1)
		perror("NBD_DO_IT");

	close(nbd);
	close(sock[0]);
	exit(0);

error:
	close(nbd);
	return 1;
}

void disconnect_nbd(int signo)
{
	if (ioctl(nbd, NBD_DISCONNECT) == -1)
		perror("NBD_DISCONNECT");
}

void service_requests(int client)
{
	struct nbd_request request;
	struct nbd_reply reply;
	off_t offset;
	size_t length;

	signal(SIGINT,  disconnect_nbd);
	signal(SIGHUP,  disconnect_nbd);
	signal(SIGTERM, disconnect_nbd);
	signal(SIGQUIT, disconnect_nbd);

	while ((length = read(client, &request, sizeof(request))) == sizeof(request))
	{
		if (request.magic != htonl(NBD_REQUEST_MAGIC))
		{
			fprintf(stderr, "[vmdk-loop] Bad header from kernel: %08x\n", request.magic);
			continue;
		}

		offset = NTOHLL(request.from);
		offset += disk_base;
		length = ntohl(request.len);

		if (request.type == htonl(NBD_SHUTDOWN))
			break;
		else if (request.type == htonl(NBD_WRITE))
			write_request(offset, length, client);

		reply.magic = htonl(NBD_REPLY_MAGIC);
		reply.error = 0;
		memcpy(reply.handle, request.handle, 8);

		if (write(client, &reply, sizeof(reply)) != sizeof(reply))
		{
			perror("write to nbd");
			continue;
		}

		if (request.type == htonl(NBD_READ))
			read_request(offset, length, client);
	}

	if (length == -1)
		perror("read from nbd");
}

int main(int argc, char *argv[])
{
	char *filename, *nbdev;
	unsigned int partition;
	int client, allow_write = 0;

	if ((argc >= 2) && !strcmp(argv[1], "-w"))
	{
		allow_write = 1;
		argc--;
		argv++;
	}

	if (argc < 3)
	{
		fprintf(stderr, "Usage: vmdk-loop [-w] <network block device> <virtual disk> [partition number]\n");
		return 1;
	}

	nbdev = argv[1];
	filename = argv[2];
	partition = (argc >= 4) ? atoi(argv[3]) : 0;

	if (allow_write)
		printf("WARNING: Enabling experimental write access\n");

	if (open_disk(filename, allow_write) == -1)
		return 1;

	if (read_partition_info(partition) == -1)
		goto error;

	if ((client = connect_nbd(nbdev, disk_limit - disk_base)) == -1)
		goto error;

	printf("%s now connected to %s", nbdev, filename);
	if (partition != 0)
		printf(" (partition %d)", partition);
	printf(": press Ctrl-C when done.\n");

	service_requests(client);

	close(client);
	close(disk_fd);
	return 0;

error:
	close(disk_fd);
	return 1;
}

